#ifndef __GURUMDDS_DCPSX__
#define __GURUMDDS_DCPSX__

#ifdef __cplusplus
extern "C" {
#endif

#if defined(__GNUC__) || defined(__clang__)
#define GURUMDDS_DEPRECATED(x) __attribute__((deprecated(x)))
#elif defined(_MSC_VER)
#define GURUMDDS_DEPRECATED(x) __declspec(deprecated(x))
#else
#define GURUMDDS_DEPRECATED(x)
#endif

/**
 * @file
 * GurumDDS Vendor Extension APIs
 */

#include <stdbool.h>

#define dds_XCDR_LZ4_DATA_REPRESENTATION ((int16_t)0x8001)

typedef struct _dds_StringProperty {
	char* key;
	void* value;
} dds_StringProperty;

typedef struct {
	uint8_t prefix[12];
	uint32_t entityId;
} dds_GUID_t;

/** The dds_SampleInfoEx can be used with dds_SampleInfoSeq APIs */
typedef struct {
	dds_SampleInfo info;
	dds_GUID_t src_guid;
	uint64_t seq;
	uint8_t keyhash[16];
	dds_Time_t reception_timestamp;
} dds_SampleInfoEx;

#define KEY_DCPS_DATAWRITER_DATA_SHARING_QOS "dcps.datawriter.data_sharing_qos"
#define KEY_DCPS_DATAWRITER_PUBLISH_MODE_QOS "dcps.datawriter.publish_mode_qos"
#define KEY_DCPS_DATAWRITER_PUBLISH_MODE_QOS_KIND "dcps.datawriter.publish_mode_qos.kind"
#define DCPS_DATAWRITER_PUBLISH_MODE_QOS_KIND_SYNC_MODE "dds_SYNC_PUBLISH_MODE"
#define DCPS_DATAWRITER_PUBLISH_MODE_QOS_KIND_ASYNC_MODE "dds_ASYNC_PUBLISH_MODE"
#define DCPS_DATAWRITER_PUBLISH_MODE_QOS_KIND_AUTO_MODE "dds_AUTO_PUBLISH_MODE"
#define KEY_DCPS_DATAWRITER_PUBLISH_MODE_QOS_THRESHOLD_PAYLOAD_SIZE "dcps.datawriter.publish_mode_qos.threshold_payload_size"
#define KEY_DCPS_DATAWRITER_LZ4_ACCELERATION "dcps.datawriter.lz4_compress.acceleration"

typedef enum {
	dds_DISABLE_DATASHARING,
	dds_AUTO_DATASHARING,
	dds_EXPLICIT_DATASHARING,
} dds_DataSharingQosPolicyKind;

/** The DataSharingQosPolicy can be used with dds_Publisher_create_datawriter_w_props API */
typedef struct {
	dds_DataSharingQosPolicyKind kind;
	uint64_t max_samples;
	uint64_t max_size_per_sample;
} dds_DataSharingQosPolicy;

typedef enum {
	dds_SYNC_PUBLISH_MODE,
	dds_ASYNC_PUBLISH_MODE,
	dds_AUTO_PUBLISH_MODE
} dds_PublishModeQosPolicyKind;

/** The dds_PublishModeQosPolicy can be used with dds_Publisher_create_datawriter_w_props API */
typedef struct {
	dds_PublishModeQosPolicyKind kind;
	uint32_t threshold_payload_size;
} dds_PublishModeQosPolicy;

typedef struct dds_NetworkInterfaceInfo {
	char name[256];
	char ip[256];
} dds_NetworkInterfaceInfo;

typedef void (*dds_DomainParticipant_on_remote_participant_changed)(const dds_DomainParticipant* the_participant, const dds_ParticipantBuiltinTopicData* data, dds_InstanceHandle_t handle);
typedef void (*dds_DomainParticipant_on_remote_publication_changed)(const dds_DomainParticipant* the_participant, const dds_PublicationBuiltinTopicData* data, dds_InstanceHandle_t handle);
typedef void (*dds_DomainParticipant_on_remote_subscription_changed)(const dds_DomainParticipant* the_participant, const dds_SubscriptionBuiltinTopicData* data, dds_InstanceHandle_t handle);

typedef void (*dds_DomainParticipant_on_network_changed)(const dds_DomainParticipant* the_participant, const dds_NetworkInterfaceInfo* data);

/**
 * Get a value from dds_StringProperty
 *
 * @param props properties
 * @param key key
 *
 * @return value
 */
GURUMDDS_C_DLLAPI void* dds_StringProperty_get(dds_StringProperty* props, char* key);

/**
 * Check whether propery array has a given key
 *
 * @param props properties
 * @param key key
 *
 * @return key is contained or not
 */
GURUMDDS_C_DLLAPI bool dds_StringProperty_contains(dds_StringProperty* props, char* key);

/**
* Change middleware log level (overrides the value set by config)
*
* @param loglevel minimum log level from 0(VERBOSE) to 5(FATAL)
*/
GURUMDDS_C_DLLAPI void dds_DomainParticipantFactory_set_loglevel(int loglevel/*0(VERBOSE) ~ 5(FATAL)*/);

/**
 * Get a log level from middleware
 *
 * @return minimum log level from 0(VERBOSE) to 5(FATAL)
 */
GURUMDDS_C_DLLAPI int dds_DomainParticipantFactory_get_loglevel(void);

/**
 * @internal
 * Dump all of the contained entities by middleware to standard output
 */
GURUMDDS_C_DLLAPI void dds_DomainParticipantFactory_dump(void);

/**
 * @internal
 * Dump all of the contained entities by middleware to newly allocated string
 *
 * @param self participant factory object
 *
 * @return newly allocated string
 */
GURUMDDS_C_DLLAPI char* dds_DomainParticipantFactory_dump_as_string(dds_DomainParticipantFactory* self);

/**
 * Creates a new DomainParticipant object with given participant id
 *
 * The DomainParticipant signifies that the calling application intends to join the Domain identified by the domain_id argument
 *
 * @param self domainparticipantfactory
 * @param domain_id domain id
 * @param participant_id participant id
 * @param qos domainparticipant qos
 * @param a_listener domainparticipant listener
 * @param mask domainparticipant listener mask
 *
 * @return newly created DomainParticipant
 */
GURUMDDS_C_DLLAPI dds_DomainParticipant* dds_DomainParticipantFactory_create_participant_w_id(dds_DomainParticipantFactory* self, dds_DomainId_t domainId, uint16_t participantId, const dds_DomainParticipantQos* qos, const dds_DomainParticipantListener* a_listener, dds_StatusMask mask);

/**
 * Returns the supported properties of DomainParticipant
 *
 * @param self participant factory object
 * @param props supported properties
 * @param props_count count of supported properties
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DomainParticipantFactory_get_supported_participant_props(dds_DomainParticipantFactory* self, const char* const* props[], uint32_t* props_count);

/**
 * Creates a new DomainParticipant object with given properties
 *
 * The DomainParticipant signifies that the calling application intends to join the Domain identified by the domain_id argument
 *
 * @param self domainparticipantfactory
 * @param domain_id domain id
 * @param qos domainparticipant qos
 * @param a_listener domainparticipant listener
 * @param mask domainparticipant listener mask
 * @param props null-terminated property array
 *
 * @return newly created DomainParticipant
 */
GURUMDDS_C_DLLAPI dds_DomainParticipant* dds_DomainParticipantFactory_create_participant_w_props(dds_DomainParticipantFactory* self, dds_DomainId_t domainId, const dds_DomainParticipantQos* qos, const dds_DomainParticipantListener* a_listener, dds_StatusMask mask, dds_StringProperty* props);

/**
 * Get a context variable from internal entity storage
 *
 * @param self entity
 * @param index storage index
 *
 * @return context variable
 */
GURUMDDS_C_DLLAPI void* dds_Entity_get_context(dds_Entity* self, int32_t index);

/**
 * Set a context variable into internal entity storage
 *
 * @param self entity
 * @param index storage index
 * @param context context variable
 *
 * @return true on succeed, false on out-of-storage
 */
GURUMDDS_C_DLLAPI bool dds_Entity_set_context(dds_Entity* self, int32_t index, void* context);

/**
 * Retrieves the GUID associated with the Entity
 *
 * @param self entity
 * @param guid guid
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_Entity_get_guid(dds_Entity* self, dds_GUID_t* guid);

// Internal Qos Profile APIs -----------------------------------------------------------------------
// dds_ReturnCode_t dds_DomainParticipantFactoryQos_from_profile(const char* profile_path, dds_DomainParticipantFactoryQos* qos);
// dds_ReturnCode_t dds_DomainParticipantQos_from_profile(const char* profile_path, dds_DomainParticipantQos* qos);
// dds_ReturnCode_t dds_TopicQos_from_profile(const char* profile_path, dds_TopicQos* qos);
// dds_ReturnCode_t dds_PublisherQos_from_profile(const char* profile_path, dds_PublisherQos* qos);
// dds_ReturnCode_t dds_SubscriberQos_from_profile(const char* profile_path, dds_SubscriberQos* qos);
// dds_ReturnCode_t dds_DataWriterQos_from_profile(const char* profile_path, dds_DataWriterQos* qos);
// dds_ReturnCode_t dds_DataReaderQos_from_profile(const char* profile_path, dds_DataReaderQos* qos);

// dds_ReturnCode_t dds_DomainParticipantFactoryQos_create(dds_DomainParticipantFactoryQos* qos);
// dds_ReturnCode_t dds_DomainParticipantQos_create(dds_DomainParticipantQos* qos);
// dds_ReturnCode_t dds_PublisherQos_create(dds_PublisherQos* qos);
// dds_ReturnCode_t dds_SubscriberQos_create(dds_SubscriberQos* qos);
// dds_ReturnCode_t dds_TopicQos_create(dds_TopicQos* qos);
// dds_ReturnCode_t dds_DataWriterQos_create(dds_DataWriterQos* qos);
// dds_ReturnCode_t dds_DataReaderQos_create(dds_DataReaderQos* qos);

// Listener Extensions -----------------------------------------------------------------------------
// getter: return associated listener context from entity
// setter: associate new listener context and returns previous one

/**
 * load listener context from domain participant
 *
 * @param self domain participant
 *
 * @return listener context
 */
GURUMDDS_C_DLLAPI void* dds_DomainParticipant_get_listener_context(dds_DomainParticipant* self);

/**
 * store listener context to domain participant
 *
 * @param self domain participant
 * @param context listener context
 *
 * @return previously stored listener context
 */
GURUMDDS_C_DLLAPI void* dds_DomainParticipant_set_listener_context(dds_DomainParticipant* self, void* context);

/**
 * load listener context from topic
 *
 * @param self topic
 *
 * @return listener context
 */
GURUMDDS_C_DLLAPI void* dds_Topic_get_listener_context(dds_Topic* self);

/**
 * store listener context to topic
 *
 * @param self topic
 * @param context listener context
 *
 * @return previously stored listener context
 */
GURUMDDS_C_DLLAPI void* dds_Topic_set_listener_context(dds_Topic* self, void* context);

/**
 * load listener context from publisher
 *
 * @param self publisher
 *
 * @return listener context
 */
GURUMDDS_C_DLLAPI void* dds_Publisher_get_listener_context(dds_Publisher* self);

/**
 * store listener context to publisher
 *
 * @param self publisher
 * @param context listener context
 *
 * @return previously stored listener context
 */
GURUMDDS_C_DLLAPI void* dds_Publisher_set_listener_context(dds_Publisher* self, void* context);

/**
 * load listener context from datawriter
 *
 * @param self datawriter
 *
 * @return listener context
 */
GURUMDDS_C_DLLAPI void* dds_DataWriter_get_listener_context(dds_DataWriter* self);

/**
 * store listener context to datawriter
 *
 * @param self datawriter
 * @param context listener context
 *
 * @return previously stored listener context
 */
GURUMDDS_C_DLLAPI void* dds_DataWriter_set_listener_context(dds_DataWriter* self, void* context);

/**
 * load listener context from subscriber
 *
 * @param self subscriber
 *
 * @return listener context
 */
GURUMDDS_C_DLLAPI void* dds_Subscriber_get_listener_context(dds_Subscriber* self);

/**
 * store listener context to subscriber
 *
 * @param self subscriber
 * @param context listener context
 *
 * @return previously stored listener context
 */
GURUMDDS_C_DLLAPI void* dds_Subscriber_set_listener_context(dds_Subscriber* self, void* context);

/**
 * load listener context from datareader
 *
 * @param self datareader
 *
 * @return listener context
 */
GURUMDDS_C_DLLAPI void* dds_DataReader_get_listener_context(dds_DataReader* self);

/**
 * store listener context to datareader
 *
 * @param self datareader
 * @param context listener context
 *
 * @return previously stored listener context
 */
GURUMDDS_C_DLLAPI void* dds_DataReader_set_listener_context(dds_DataReader* self, void* context);

// Entity naming Extensions ------------------------------------------------------------------------
/**
 * Name the domain participant
 *
 * @param self domain participant
 * @param name name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DomainParticipant_get_name(dds_DomainParticipant* self, char* name/*[256]*/);

/**
 * Get domain participant's name
 *
 * @param self domain participant
 * @param name buffer to store name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DomainParticipant_set_name(dds_DomainParticipant* self, char* name/*[256]*/);

/**
 * Name the publisher
 *
 * @param self publisher
 * @param name name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_Publisher_get_name(dds_Publisher* self, char* name/*[256]*/);

/**
 * Get publisher's name
 *
 * @param self publisher
 * @param name buffer to store name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_Publisher_set_name(dds_Publisher* self, char* name/*[256]*/);

/**
 * Name the datawriter
 *
 * @param self datawriter
 * @param name name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataWriter_get_name(dds_DataWriter* self, char* name/*[256]*/);

/**
 * Get datawriter's name
 *
 * @param self datawriter
 * @param name buffer to store name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataWriter_set_name(dds_DataWriter* self, char* name/*[256]*/);

/**
 * Name the subscriber
 *
 * @param self subscriber
 * @param name name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_Subscriber_get_name(dds_Subscriber* self, char* name/*[256]*/);

/**
 * Get subscriber's name
 *
 * @param self subscriber
 * @param name buffer to store name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_Subscriber_set_name(dds_Subscriber* self, char* name/*[256]*/);

/**
 * Name the datareader
 *
 * @param self datareader
 * @param name name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_get_name(dds_DataReader* self, char* name/*[256]*/);

/**
 * Get datareader's name
 *
 * @param self datareader
 * @param name buffer to store name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_set_name(dds_DataReader* self, char* name/*[256]*/);

/**
 * Creates a new rawly TypeSupport object
 *
 * @return newly created TypeSupport
 */
GURUMDDS_C_DLLAPI dds_TypeSupport* dds_TypeSupport_raw_create();

/**
 * Creates a new rawly TypeSupport object with key
 *
 * @return newly created TypeSupport
 */
GURUMDDS_C_DLLAPI dds_TypeSupport* dds_TypeSupport_raw_create_w_key();

/**
 * Creates a new rawly TypeSupport object with typehash and key
 *
 * @param typehash type hash
 * @param has_key type has key or not
 *
 * @return newly created TypeSupport
 */
GURUMDDS_C_DLLAPI dds_TypeSupport* dds_TypeSupport_raw_create_w_typehash(void* typehash, bool has_key);

/**
 * Destroy a TypeSupport object
 */
GURUMDDS_C_DLLAPI GURUMDDS_DEPRECATED("Use dds_TypeSupport_delete(dds_TypeSupport*) instead.") void dds_TypeSupport_destroy(dds_TypeSupport* self);

/**
 * Name the typesupport
 *
 * @param self typesupport
 * @param name name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_TypeSupport_get_name(dds_TypeSupport* self, char* name/*[256]*/);

/**
 * Get typesupport's name
 *
 * @param self typesupport
 * @param name buffer to store name
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_TypeSupport_set_name(dds_TypeSupport* self, char* name/*[256]*/);

/**
 * Serialize UserData in big endian byte order
 *
 * @param self typesupport
 * @param data userdata pointer
 * @param output_size size of serialized userdata
 *
 * @return newly allocated userdata octets
 */
GURUMDDS_C_DLLAPI void* dds_TypeSupport_serialize_be(dds_TypeSupport* self, void* data, size_t* output_size);

/**
 * Serialize UserData in little endian byte order
 *
 * @param self typesupport
 * @param data userdata pointer
 * @param output_size size of serialized userdata
 *
 * @return newly allocated userdata octets
 */
GURUMDDS_C_DLLAPI void* dds_TypeSupport_serialize_le(dds_TypeSupport* self, void* data, size_t* output_size);

/**
 * Extract metastring data from TypeSupport. The returned metastring should be freed after used
 *
 * @param self typesupport
 *
 * @return metastring
 */
GURUMDDS_C_DLLAPI char* dds_TypeSupport_extract_metastring(dds_TypeSupport* self);

/**
 * Create DynamicTypeSupport with metastring
 *
 * @param metastring metastring
 *
 * @return newly created DynamicTypeSupport
 */
GURUMDDS_C_DLLAPI dds_DynamicTypeSupport* dds_DynamicTypeSupport_create_type_support_with_metastring(char* metastring);

/**
 * Get DynamicType
 *
 * @param self typesupport
 *
 * @return DynamicType object
 */
GURUMDDS_C_DLLAPI dds_DynamicType* dds_DynamicTypeSupport_get_dynamictype(dds_DynamicTypeSupport* self);

/**
 * Check the TypeSupport has plain type (fixed type)
 *
 * @param self typesupport
 *
 * @return true when the type is plain, false otherwise
 */
GURUMDDS_C_DLLAPI bool dds_TypeSupport_is_plain(dds_TypeSupport* self);

// Entity accessing extensions ------------------------------------------------------------------------
/**
 * Enumerates all domain participant that the domain participant factory has
 *
 * @param self domain participant factory
 * @param participants sequence of domain participant
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DomainParticipantFactory_get_contained_entities(dds_DomainParticipantFactory* self, dds_InstanceHandleSeq* participants);

/**
 * Enumerates all publisher, subscriber, topic and typesupport that the domain participant has (including builtin entity)
 *
 * @param self domain participant
 * @param publishers sequence of publishers
 * @param subscribers sequence of subscribers
 * @param topics sequence of topics
 * @param typesupports sequence of typesupports
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DomainParticipant_get_contained_entities(dds_DomainParticipant* self, dds_InstanceHandleSeq* publishers, dds_InstanceHandleSeq* subscribers, dds_InstanceHandleSeq* topics, dds_InstanceHandleSeq* typesupports);

/**
 * Enumerates all datawriter that the publisher has (including builtin entity)
 *
 * @param self publisher
 * @param participants sequence of datawriter
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_Publisher_get_contained_entities(dds_Publisher* self, dds_InstanceHandleSeq* datawriters);

/**
 * Enumerates all domain participant that the domain participant factory has (including builtin entity)
 *
 * @param self subscriber
 * @param participants sequence of datareader
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_Subscriber_get_contained_entities(dds_Subscriber* self, dds_InstanceHandleSeq* datareaders);

/**
 * Returns the domain participant with the given name
 *
 * @param self domain participant factory
 * @param name name of domain participant
 *
 * @return domain participant
 */
GURUMDDS_C_DLLAPI dds_DomainParticipant* dds_DomainParticipantFactory_get_contained_participant_by_name(dds_DomainParticipantFactory* self, char* name);

/**
 * Returns the publisher with the given name
 *
 * @param self domain participant
 * @param name name of publisher
 *
 * @return publisher
 */
GURUMDDS_C_DLLAPI dds_Publisher* dds_DomainParticipant_get_contained_publisher_by_name(dds_DomainParticipant* self, char* name);

/**
 * Returns the subscriber with the given name
 *
 * @param self domain participant
 * @param name name of subscriber
 *
 * @return subscriber
 */
GURUMDDS_C_DLLAPI dds_Subscriber* dds_DomainParticipant_get_contained_subscriber_by_name(dds_DomainParticipant* self, char* name);

/**
 * Returns the topic with the given name
 *
 * @param self domain participant
 * @param name name of topic
 *
 * @return topic
 */
GURUMDDS_C_DLLAPI dds_Topic* dds_DomainParticipant_get_contained_topic_by_name(dds_DomainParticipant* self, char* name);

/**
 * Returns the typesupport with the given name
 *
 * @param self domain participant
 * @param name name of typesupport
 *
 * @return typesupport
 */
GURUMDDS_C_DLLAPI dds_TypeSupport* dds_DomainParticipant_get_contained_typesupport_by_name(dds_DomainParticipant* self, char* name);

/**
 * Returns the typesupport with the given type name
 *
 * @param self domain participant
 * @param type_name type name of typesupport
 *
 * @return typesupport
 */
GURUMDDS_C_DLLAPI dds_TypeSupport* dds_DomainParticipant_get_contained_typesupport_by_type_name(dds_DomainParticipant* self, char* type_name);

/**
 * Find DynamicTypeSupports by type_name using builtin discovery data.
 * GURUMDDS_USE_VENDOR_SPECIFIC_PARAMETER must be allowed on the other side
 * PID_GURUMDDS_CDRMETA (0x8001) on Data(w) and Data(r) is used to create DynamicTypeSupport
 *
 * @param self domain participant
 * @param type_name type name trying to find
 *
 * @return dds_DataSeq object contains dds_DynamicTypeSupport objects
 */
GURUMDDS_C_DLLAPI dds_DataSeq* dds_DomainParticipant_lookup_delivered_types(dds_DomainParticipant* self, const char* type_name);

/**
 * Returns the datawriter with the given name
 *
 * @param self publisher
 * @param name name of datawriter
 *
 * @return datawriter
 */
GURUMDDS_C_DLLAPI dds_DataWriter* dds_Publisher_get_contained_datawriter_by_name(dds_Publisher* self, char* name);

/**
 * Returns the datareader with the given name
 *
 * @param self subscriber
 * @param name name of datareader
 *
 * @return datareader
 */
GURUMDDS_C_DLLAPI dds_DataReader* dds_Subscriber_get_contained_datareader_by_name(dds_Subscriber* self, char* name);

/**
 * Returns the typesupport of the datawriter
 *
 * @param self datawriter
 *
 * @return typesupport
 */
GURUMDDS_C_DLLAPI dds_TypeSupport* dds_DataWriter_get_typesupport(dds_DataWriter* self);

/**
 * Returns the typesupport of the datareader
 *
 * @param self datareader
 *
 * @return typesupport
 */
GURUMDDS_C_DLLAPI dds_TypeSupport* dds_DataReader_get_typesupport(dds_DataReader* self);

/**
 * Read a collection of Data values from the DataReader
 *
 * Deliver dds_SampleInfoEx instead of dds_SampleInfo
 *
 * The size of the returned collection will be limited to the specified max_samples
 *
 * If the DataReader has no samples that meet the constraints, the return value will be NO_DATA
 *
 * @param self datareader
 * @param data_values data sequence
 * @param sample_infos extended sample info sequence
 * @param max_samples max samples
 * @param sample_states sample states
 * @param view_states view states
 * @param instance_states instance states
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_read_w_sampleinfoex(dds_DataReader* self, dds_DataSeq* data_values, dds_SampleInfoSeq* sample_infos, int32_t max_samples, dds_SampleStateMask sample_states, dds_ViewStateMask view_states, dds_InstanceStateMask instance_states);

/**
 * Take a collection of Data values from the DataReader
 *
 * Deliver dds_SampleInfoEx instead of dds_SampleInfo
 *
 * The size of the returned collection will be limited to the specified max_samples
 *
 * If the DataReader has no samples that meet the constraints, the return value will be NO_DATA
 *
 * @param self datareader
 * @param data_values data sequence
 * @param sample_infos extended sample info sequence
 * @param max_samples max samples
 * @param sample_states sample states
 * @param view_states view states
 * @param instance_states instance states
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_take_w_sampleinfoex(dds_DataReader* self, dds_DataSeq* data_values, dds_SampleInfoSeq* sample_infos, int32_t max_samples, dds_SampleStateMask sample_states, dds_ViewStateMask view_states, dds_InstanceStateMask instance_states);

// Time extensions ------------------------------------------------------------------------
/**
 * Get current wall-clock time
 *
 * @param current_time wall-clock time
 */
GURUMDDS_C_DLLAPI void dds_Time_get_current_time(dds_Time_t* current_time);

/**
 * Sleep the current thread by the given amount of time
 *
 * @param sleep_time sleep time
 */
GURUMDDS_C_DLLAPI void dds_Time_sleep(const dds_Time_t* sleep_time);

/**
 * Gets the time difference of two time objects
 *
 * @param lhs time object (left hand side)
 * @param rhs time object (right hand side)
 * @param diff time difference
 */
GURUMDDS_C_DLLAPI void dds_Time_diff(dds_Time_t* lhs, dds_Time_t* rhs, dds_Time_t* diff);

/**
 * Performs the same function as raw_write except that it also provides the value for the seq that is made available to DataWriter objects by means of the seq attribute inside the SampleInfoEx
 *
 * @see dds_DataWriter_raw_write
 *
 * @param self datawriter
 * @param instance_data instance data
 * @param sampleinfoex sample info extension
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataWriter_write_w_sampleinfoex(dds_DataWriter* self, const void* instance_data, const dds_SampleInfoEx* sampleinfoex);

// Raw-Data extensions -----------------------------------------------------------------------------
/**
 * Writes the value of a raw data. When this operation is used, the Service will automatically supply the value of the source_timestamp that is made available to DataReader objects by means of the source_timestamp attribute
 *
 * As a side effect, this operation asserts liveliness on the DataWriter itself, the Publisher and the DomainParticipant
 *
 * @param self datawriter
 * @param raw_data raw data
 * @param raw_data_size raw data size
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataWriter_raw_write(dds_DataWriter* self, const void* raw_data, uint32_t raw_data_size);

/**
 * Performs the same function as raw_write except that it also provides the value for the source_timestamp that is made available to DataReader objects by means of the source_timestamp attribute inside the SampleInfo
 *
 * @see dds_DataWriter_raw_write
 *
 * @param self datawriter
 * @param raw_data raw data
 * @param raw_data raw data size
 * @param source_timestamp source timestamp
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataWriter_raw_write_w_timestamp(dds_DataWriter* self, const void* raw_data, uint32_t raw_data_size, const dds_Time_t* source_timestamp);

/**
 * Performs the same function as raw_write except that it also provides the value for the seq that is made available to DataWriter objects by means of the seq attribute inside the SampleInfoEx
 *
 * @see dds_DataWriter_raw_write
 *
 * @param self datawriter
 * @param raw_data raw data
 * @param raw_data raw data size
 * @param sampleinfoex sample info extension
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataWriter_raw_write_w_sampleinfoex(dds_DataWriter* self, const void* raw_data, uint32_t raw_data_size, dds_SampleInfoEx* sampleinfoex);

/**
 * Read a collection of raw data values from the DataReader
 *
 * The size of the returned collection will be limited to the specified max_samples
 *
 * If the DataReader has no samples that meet the constraints, the return value will be NO_DATA
 *
 * @param self datareader
 * @param raw_data_values raw data sequence
 * @param sample_infos sample info sequence
 * @param raw_data_sizes raw data size sequence
 * @param max_samples max samples
 * @param sample_states sample states
 * @param view_states view states
 * @param instance_states instance states
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_raw_read(dds_DataReader* self, dds_InstanceHandle_t handle, dds_DataSeq* raw_data_values, dds_SampleInfoSeq* sample_infos, dds_UnsignedLongSeq* raw_data_sizes, int32_t max_samples, dds_SampleStateMask sample_states, dds_ViewStateMask view_states, dds_InstanceStateMask instance_states);

/**
 * Performs the same function as raw_read except that it provides the sequence for dds_SampleInfoEx instead of dds_SampleInfo
 *
 * @see dds_DataReader_raw_read
 *
 * @param self datareader
 * @param raw_data_values raw data sequence
 * @param sample_infoexs sample info extension sequence
 * @param raw_data_sizes raw data size sequence
 * @param max_samples max samples
 * @param sample_states sample states
 * @param view_states view states
 * @param instance_states instance states
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_raw_read_w_sampleinfoex(dds_DataReader* self, dds_InstanceHandle_t handle, dds_DataSeq* raw_data_values, dds_SampleInfoSeq* sample_infoexs, dds_UnsignedLongSeq* raw_data_sizes, int32_t max_samples, dds_SampleStateMask sample_states, dds_ViewStateMask view_states, dds_InstanceStateMask instance_states);

/**
 * Take a collection of raw data values from the DataReader
 *
 * The size of the returned collection will be limited to the specified max_samples
 *
 * If the DataReader has no samples that meet the constraints, the return value will be NO_DATA
 *
 * @param self datareader
 * @param raw_data_values raw data sequence
 * @param sample_infos sample info sequence
 * @param raw_data_sizes raw data size sequence
 * @param max_samples max samples
 * @param sample_states sample states
 * @param view_states view states
 * @param instance_states instance states
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_raw_take(dds_DataReader* self, dds_InstanceHandle_t handle, dds_DataSeq* raw_data_values, dds_SampleInfoSeq* sample_infos, dds_UnsignedLongSeq* raw_data_sizes, int32_t max_samples, dds_SampleStateMask sample_states, dds_ViewStateMask view_states, dds_InstanceStateMask instance_states);

/**
 * Performs the same function as raw_take except that it provides the sequence for dds_SampleInfoEx instead of dds_SampleInfo
 *
 * @see dds_DataReader_raw_take
 *
 * @param self datareader
 * @param raw_data_values data sequence
 * @param sample_infoexs sample info extension sequence
 * @param raw_data_sizes raw data size sequence
 * @param max_samples max samples
 * @param sample_states sample states
 * @param view_states view states
 * @param instance_states instance states
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_raw_take_w_sampleinfoex(dds_DataReader* self, dds_InstanceHandle_t handle, dds_DataSeq* raw_data_values, dds_SampleInfoSeq* sample_infoexs, dds_UnsignedLongSeq* raw_data_sizes, int32_t max_samples, dds_SampleStateMask sample_states, dds_ViewStateMask view_states, dds_InstanceStateMask instance_states);

/**
 * Returns the 'loan' (i.e. raw_data_values, and sample_infos)
 *
 * @param self datareader
 * @param raw_data_values raw data sequence
 * @param sample_infos sample info sequence
 * @param raw_data_sizes raw data size sequence
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_raw_return_loan(dds_DataReader* self, dds_DataSeq* raw_data_values, dds_SampleInfoSeq* sample_infos, dds_UnsignedLongSeq* raw_data_sizes);

// Shared memory sharing -----------------------------------------------------------------------------
/**
 * Checks the consistency of data received through Data Sharing
 * Returns true if the corresponding Writer has not yet overwritten the data, and false if it has been overwritten
 *
 * @param self datareader
 * @param raw_data raw data
 * @param info sample info externsion
 *
 * @return data is consistent or not
 */
GURUMDDS_C_DLLAPI bool dds_DataReader_is_data_consistent(dds_DataReader* self, const void* raw_data, const dds_SampleInfoEx* info);

/**
 * Get the GUID associated with the DataWriter
 *
 * @param self datawriter
 * @param guid guid
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataWriter_get_guid(dds_DataWriter* self, uint8_t* guid);

/**
 * Get the GUID associated with the DataWriter via publication handle
 *
 * @param self datareader
 * @param publication_handle publication handle
 * @param guid guid
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_get_guid_from_publication_handle(dds_DataReader* self, dds_InstanceHandle_t publication_handle, uint8_t* guid);

/**
 * Get the GUID associated with the DataReader via subscription handle
 *
 * @param self datawriter
 * @param subscriptuion_handle publication handle
 * @param guid guid
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataWriter_get_guid_from_subscription_handle(dds_DataWriter* self, dds_InstanceHandle_t subscription_handle, uint8_t* guid);

/**
 * Returns a string representation of the ReturnCode
 *
 * @param retcode return code
 *
 * @return string
 */
GURUMDDS_C_DLLAPI dds_String dds_ReturnCode_to_string(dds_ReturnCode_t retcode);

/**
 * Allocates memory and returns a pointer to the allocated memory
 * The Memory is not initialized
 *
 * @param n size
 *
 * @return allocated memory pointer
 */
GURUMDDS_C_DLLAPI void* dds_malloc(size_t n);

/**
 * Allocates memory and returns a pointer to the allocated memory
 * The Memory is initialized
 *
 * @param nmemb array size
 * @param n element size
 *
 * @return allocated memory pointer
 */
GURUMDDS_C_DLLAPI void* dds_calloc(size_t nmemb, size_t s);

/**
 * Returns new memory for a duplicated string
 *
 * @param s string to duplicate
 *
 * @return allocated memory pointer
 */
GURUMDDS_C_DLLAPI char* dds_strdup(const char* s);

/**
 * Returns new memory for a duplicated UTF-16 string
 *
 * @param s UTF-16 string to duplicate
 *
 * @return allocated memory pointer
 */
GURUMDDS_C_DLLAPI uint16_t* dds_wcsdup(const uint16_t* s);

/**
 * Frees the memory space to by pointer
 *
 * @param memory pointer
 */
GURUMDDS_C_DLLAPI void dds_free(void* ptr);

/**
 * Returns the UTF-16 string converted from wide string
 *
 * @param UTF-16 string
 */
GURUMDDS_C_DLLAPI dds_Wstring dds_wide_to_utf16(const wchar_t* s);

/**
 * Returns the wide string converted from UTF-16 string
 *
 * @param wide string
 */
GURUMDDS_C_DLLAPI wchar_t* dds_utf16_to_wide(const dds_Wchar* s);

/**
 * @internal
 * Dump all of the contained policy by QoS to newly allocated string
 *
 * @param self QoS object
 *
 * @return newly allocated string
 */
GURUMDDS_C_DLLAPI char* dds_DomainParticipantQos_dump_as_string(dds_DomainParticipantQos* qos);

/** @see dds_DomainParticipantQos_dump_as_string */
GURUMDDS_C_DLLAPI char* dds_TopicQos_dump_as_string(dds_TopicQos* qos);

/** @see dds_DomainParticipantQos_dump_as_string */
GURUMDDS_C_DLLAPI char* dds_PublisherQos_dump_as_string(dds_PublisherQos* qos);

/** @see dds_DomainParticipantQos_dump_as_string */
GURUMDDS_C_DLLAPI char* dds_SubscriberQos_dump_as_string(dds_SubscriberQos* qos);

/** @see dds_DomainParticipantQos_dump_as_string */
GURUMDDS_C_DLLAPI char* dds_DataWriterQos_dump_as_string(dds_DataWriterQos* qos);

/** @see dds_DomainParticipantQos_dump_as_string */
GURUMDDS_C_DLLAPI char* dds_DataReaderQos_dump_as_string(dds_DataReaderQos* qos);

/**
 * Get static discovery's enable status
 *
 * @return status static discovery enable
 **/
GURUMDDS_C_DLLAPI bool dds_DomainParticipantFactory_static_discovery_enabled(dds_DomainParticipantFactory* self);

/**
 * Generate typehash from metastring
 *
 * @param participant domain participant object
 * @param mstastring metastring
 * @param typehash typehash holder (at least 16 bytes long)
 *
 * @return true on succeed, or false on failed
 **/
GURUMDDS_C_DLLAPI bool dds_DomainParticipant_generate_typehash(dds_DomainParticipant* self, const char* metastring, void* typehash);

/**
 * Get data from DynamicData
 *
 * @param self dynamic data
 *
 * @return data
 **/
GURUMDDS_C_DLLAPI void* dds_DynamicData_get_data(dds_DynamicData* self);

/**
 * Creates a DataWriter with given properties
 *
 * The returned DataWriter will be attached and belongs to the Publisher. The DataWriter returned by the create_datawriter operation will in fact be a derived class, specific to the data-type associated with the Topic
 *
 * @param self publisher
 * @param a_topic topic
 * @param qos datawriter qos
 * @param a_listener datawriter listener
 * @param mask datawriter listener mask
 * @param props null-terminated property array
 *
 * @return datawriter
 */
GURUMDDS_C_DLLAPI dds_DataWriter* dds_Publisher_create_datawriter_w_props(dds_Publisher* self, const dds_Topic* a_topic, const dds_DataWriterQos* qos, const dds_DataWriterListener* a_listener, dds_StatusMask mask, dds_StringProperty* props);

// Zero copy API
/**
 * Borrows the shared memory created via DataSharing
 *
 * If you wish to return the borrowed pointer without calling raw_write, please call the return_loaned_sample function
 *
 * @param self datawriter
 * @param data data to store the borrowed pointer
 * @param size size to store the borrowed size
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataWriter_get_loan(dds_DataWriter* self, void** data, size_t* size);

/**
 * Returns the 'loan' that shared memory borrowed through the get_loan function
 *
 * This function must not be used to return shared memory that was passed through raw_write
 *
 * @param self datawriter
 * @param data loaned data
 */
GURUMDDS_C_DLLAPI void dds_DataWriter_return_loaned_sample(dds_DataWriter* self, void* data);

/**
 * Checks if DataSharing is enabled
 *
 * @param self datawriter
 *
 * @return DataSharing is enabled or not
 */
GURUMDDS_C_DLLAPI bool dds_DataWriter_is_data_sharing_enabled(dds_DataWriter* self);

/** This operation is not supported */
GURUMDDS_C_DLLAPI bool dds_DataReader_is_data_sharing_enabled(dds_DataReader* self);

/**
 * Loans the next, non-previously accessed Data value from the DataReader
 *
 * The operation also copies the corresponding SampleInfoEx. The implied order among the samples stored in the DataReader is the same as for the read operation
 *
 * @param self datareader
 * @param data_value data to store the borrowed pointer
 * @param sample_info sample info extensions
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_read_next_loaned_sample(dds_DataReader* self, void** data_value, dds_SampleInfoEx* sample_info);

/**
 * Loans the next, non-previously accessed Data value from the DataReader
 *
 * The operation also copies the corresponding SampleInfoEx. The implied order among the samples stored in the DataReader is the same as for the read operation
 *
 * @param self datareader
 * @param data_value data to store the borrowed pointer
 * @param sample_info sample info extensions
 *
 * @return standard return code
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_take_next_loaned_sample(dds_DataReader* self, void** data_value, dds_SampleInfoEx* sample_info);

/** @see dds_DataReader_take_next_sample */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_take_next_sample_w_info_ex(dds_DataReader* self, void* data_value, dds_SampleInfoEx* sample_info);

/**
 * Returns the 'loan' that shared memory borrowed through the loaned sample function
 *
 * @param self datawriter
 * @param data_value loaned data
 */
GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DataReader_return_loaned_sample(dds_DataReader* self, void* data_value);

typedef struct dds_md5_ctx dds_md5_ctx;

/**
 * Allocate memory for a new MD5 context
 * 
 * This function is the first fucntion to call before initializing the MD5 context with dds_md5_init()
 * 
 * @return pointer to the allocated MD5 context, or NULL if memoery allocation fails
 */
GURUMDDS_C_DLLAPI void* dds_md5_ctx_alloc();

/**
 * Initalize the MD5 context
 * 
 * Prepares the MD5 context for hashing new data
 * This function is must be called before any calls to dds_md5_update()
 * 
 * @param ctx pointer to the MD5 context to initialize
 */
GURUMDDS_C_DLLAPI void dds_md5_init(dds_md5_ctx* ctx);

/**
 * Update the MD5 context with new data
 * 
 * Feed data into the context for MD5 hashing
 * This function can be called repeatedly with chunks of the data. if not all data is available at once
 * 
 * @param ctx pointer to the MD5 context
 * @param data pointer to the data to hash
 * @param size Size of the data in bytes
 */
GURUMDDS_C_DLLAPI void dds_md5_update(dds_md5_ctx* ctx, const void* data, unsigned long size);

/**
 * Finalize the MD5 hash and outputs the result
 * 
 * Completes the MD5 hashing process and writes the MD5 hash to the provided result buffer
 * The MD5 context cannot be used again for hashing without reinitializing with dds_md5_init()
 * 
 * @param result buffer to store the 16-byte MD5 hash result
 * @param ctx pointer to the MD5 context
 */
GURUMDDS_C_DLLAPI void dds_md5_final(unsigned char* result, dds_md5_ctx* ctx);

/**
 * Free the memory allocated for an MD5 context
 * 
 * Releases the memory allocated for the MD5 context
 * 
 * @param ctx pointer to the MD5 context to free
 */
GURUMDDS_C_DLLAPI void dds_md5_ctx_free(dds_md5_ctx* ctx);

GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DomainParticipantFactory_load_profile_file(const char* profile_path);

GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DomainParticipantFactory_get_participant_qos_from_profile(const char* profile_name, dds_DomainParticipantQos* qos);

GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DomainParticipantFactory_get_datawriter_qos_from_profile(const char* profile_name, dds_DataWriterQos* qos);

GURUMDDS_C_DLLAPI dds_ReturnCode_t dds_DomainParticipantFactory_get_datareader_qos_from_profile(const char* profile_name, dds_DataReaderQos* qos);

GURUMDDS_C_DLLAPI dds_DomainParticipant* dds_DomainParticipantFactory_create_participant_w_profile(dds_DomainParticipantFactory* self, dds_DomainId_t domain_id, const char* profile_name, const dds_DomainParticipantListener* a_listener, dds_StatusMask mask);

GURUMDDS_C_DLLAPI dds_DataWriter* dds_Publisher_create_datawriter_w_profile(dds_Publisher* self, const dds_Topic* a_topic, const char* profile_name, const dds_DataWriterListener* a_listener, dds_StatusMask mask);

GURUMDDS_C_DLLAPI dds_DataReader* dds_Subscriber_create_datareader_w_profile(dds_Subscriber* self, const dds_Topic* a_topic, const char* profile_name, const dds_DataReaderListener* a_listener, dds_StatusMask mask);

#ifdef __cplusplus
}
#endif

#endif /* __GURUMDDS_DCPSX__ */
